"""
Modifiers
===========
When a request for configuration arrives at the control plane, it will first
retrieve data from all configured sources.

Directly after it has run all sources, it will apply modifications to the data.
If no modifiers are configured, the data is returned as-is.

Modifiers are enabled by adding the following configuration:

.. code-block:: yaml

   modifiers:
     - <modifier_name>

   global_modifiers:
     - <global_modifier_name>

The modifier name must match the name of the entry point

`todo add guide for adding entry points`

Modifiers will apply to each instance of the source data, whereas
global modifiers can apply to the entire set of source data.

Global modifiers are executed before modifiers.
"""
from typing import Callable
from sovereign.statistics import stats
from sovereign.decorators import memoize
from sovereign.schemas import SourceData, Instances, Instance
from sovereign.modifiers.lib import GlobalModifier, Modifier


@memoize(60)
@stats.timed('modifiers.apply_ms')
def modify_sources_in_place(
        source_data: SourceData,
        gmods: Callable[[Instances], GlobalModifier],
        mods: Callable[[Instance], Modifier],
    ) -> SourceData:
    """
    Runs all configured modifiers on received data from sources.
    Returns the data, with modifications applied.
    """
    for scope, instances in source_data.scopes.items():
        for g in gmods:
            global_modifier = g(instances)
            global_modifier.apply()
            source_data.scopes[scope] = global_modifier.join()

        for index, instance in enumerate(source_data.scopes[scope]):
            for m in mods:
                modifier = m(instance)
                if modifier.match():
                    source_data.scopes[scope][index] = modifier.apply()
    return source_data



from __future__ import division


def _round_to_n(sequence, n=100):
    """ Resolves issue when dividing by N results in a set of numbers that don't add up to N again

        E.g. 100 / 3 = 33
             33 * 3 = 99

        For the above example, this function returns [33, 33, 34]
     """
    if sum(sequence) == 0:
        return [0 for _ in sequence]
    if sum(sequence) != n:
        sequence[-1] = n - sum(sequence[:-1])
    return sequence


def _normalize_weights(sequence, n=100):
    total = sum(sequence)
    for item in sequence:
        try:
            yield int((item / total) * n)
        except ZeroDivisionError:
            yield 0


def fit_weights(clusters, total_weight=100):
    weights = list(_normalize_weights([cluster['weight'] for cluster in clusters], n=total_weight))
    for cluster, newly_assigned_weight in zip(clusters, _round_to_n(weights, n=total_weight)):
        cluster['weight'] = newly_assigned_weight
    return clusters



from sovereign.modifiers.lib import Modifier
from sovereign.context import template_context
from sovereign.utils import eds, templates


class Test(Modifier):
    def match(self):
        return True

    def apply(self):
        assert template_context
        assert eds
        assert templates
        return self.instance



"""
Source abstract class
---------------------
This class can be subclassed, installed as an entry point, and then
used via configuration.

`todo entry point install guide`
"""
import abc
from sovereign.logs import LOG
from sovereign.schemas import Instances


class Source(metaclass=abc.ABCMeta):
    def __init__(self, config: dict, scope: str):
        """
        :param config: arbitrary configuration which can be used by the subclass
        """
        self.logger = LOG
        self.config = config
        self.scope = scope

    def setup(self):
        """
        Optional method which is invoked prior to the Source running self.get()
        """

    @abc.abstractmethod
    def get(self) -> Instances:
        """
        Required method to retrieve data from an arbitrary source
        """



import yaml
from collections import defaultdict
from fastapi import APIRouter, Query
from fastapi.responses import JSONResponse
from fastapi.encoders import jsonable_encoder
from sovereign import discovery, config
from sovereign.discovery import DiscoveryTypes, select_template
from sovereign.context import safe_context
from sovereign.statistics import stats
from sovereign.utils.mock import mock_discovery_request
from sovereign.sources import get_instances_for_node, extract_node_key

router = APIRouter()


@router.get(
    '/xds_dump',
    summary='Displays all xDS resources as JSON'
)
async def display_config(
        xds_type: DiscoveryTypes = Query(..., title='xDS type', description='The type of request', example='clusters'),
        service_cluster: str = Query('*', title='The clients service cluster to emulate in this XDS request'),
        resource_names: str = Query('', title='Envoy Resource names to request'),
        region: str = Query(None, title='The clients region to emulate in this XDS request'),
        version: str = Query('1.11.1', title='The clients envoy version to emulate in this XDS request')
):
    ret = defaultdict(list)
    mock_request = mock_discovery_request(
        service_cluster=service_cluster,
        resource_names=resource_names,
        version=version,
        region=region
    )
    response = await discovery.response(
        request=mock_request,
        xds_type=xds_type.value
    )
    ret['resources'] += response.resources
    safe_response = jsonable_encoder(ret)
    return JSONResponse(content=safe_response)


@router.get(
    '/debug_template',
    summary='Dumps raw representation of template output before serialization/extra processing'
)
async def debug_template(
        xds_type: str = Query(..., title='xDS type', description='The type of request', example='clusters'),
        service_cluster: str = Query('*', title='The clients service cluster to emulate in this XDS request'),
        resource_names: str = Query('', title='Envoy Resource names to request'),
        region: str = Query(None, title='The clients region to emulate in this XDS request'),
        version: str = Query('1.11.1', title='The clients envoy version to emulate in this XDS request')
):
    mock_request = mock_discovery_request(
        service_cluster=service_cluster,
        resource_names=resource_names,
        version=version,
        region=region
    )
    template = select_template(mock_request, xds_type)
    context = safe_context(mock_request, template)
    context = dict(
        discovery_request=mock_request,
        host_header='debug',
        resource_names=mock_request.resources,
        **context
    )
    raw_template_content = await template(**context)
    safe_response = jsonable_encoder(raw_template_content)
    return JSONResponse(content=safe_response)


@router.get(
    '/source_dump',
    summary='Displays all sources that this Sovereign has polled as JSON'
)
def instances(
        service_cluster: str = Query('*', title='The clients service cluster to emulate in this XDS request'),
        modified: str = Query('yes',
                              title='Whether the sources should run Modifiers/Global Modifiers prior to being returned')
):
    node = mock_discovery_request(
        service_cluster=service_cluster
    ).node
    args = {
        'modify': yaml.safe_load(modified),
        'node_value': extract_node_key(node)
    }
    ret = get_instances_for_node(**args)
    safe_response = jsonable_encoder(ret)
    return JSONResponse(content=safe_response)


@router.get(
    '/config',
    summary='Display the current Sovereign configuration'
)
def show_configuration():
    safe_response = jsonable_encoder(config.show())
    return JSONResponse(content=safe_response)


@router.get(
    '/stats',
    summary='Displays all metrics emitted and their counters'
)
def show_stats():
    return JSONResponse(content=stats.emitted)



import random
from copy import deepcopy
from starlette.exceptions import HTTPException
from sovereign import config
from sovereign.schemas import DiscoveryRequest
from sovereign.utils.templates import resolve

priority_mapping = config.eds_priority_matrix
total_regions = len(config.regions)


def _upstream_kwargs(upstream, proxy_region=None, resolve_dns=True, default_region=None, hard_fail=config.dns_hard_fail) -> dict:
    try:
        ip_addresses = resolve(upstream['address']) if resolve_dns else [upstream['address']]
    except HTTPException:
        if hard_fail:
            raise
        ip_addresses = [upstream['address']]
    return {
        'addrs': ip_addresses,
        'port': upstream['port'],
        'region': default_region or upstream.get('region', 'unknown'),
        'zone': proxy_region
    }


def total_zones(endpoints: list) -> int:
    """
    Returns the true unique number of zones, taking into account
    that multiple endpoints can have the same zone name.

    - us-west-1
    - us-west-1   == 2 zones
    - us-east-1

    - us-west-1
    - us-west-2   == 3 zones
    - us-east-2
    """
    zones = {e['locality']['zone'] for e in endpoints}
    return len(zones)


def locality_lb_endpoints(upstreams, request: DiscoveryRequest = None, resolve_dns=True):
    if request is None:
        proxy_region = None
    else:
        proxy_region = request.node.locality.zone

    kw_args = [_upstream_kwargs(u, proxy_region, resolve_dns) for u in upstreams]
    ret = [lb_endpoints(**kw) for kw in kw_args]

    if total_zones(ret) == 1:
        # Pointless to do zone-aware load-balancing for a single zone
        return ret

    upstreams_copy = deepcopy(upstreams)
    while total_zones(ret) < total_regions:
        region = f'zone-padding-{total_zones(ret)}'
        try:
            upstream = upstreams_copy.pop()
        except IndexError:
            # When adding zone-padding, use a randomly selected upstream
            # However, the random selection should be consistent across control-planes
            # otherwise the version_info of the response will be constantly different
            random.seed(128)
            upstream = random.choice(upstreams)
        kw_args = _upstream_kwargs(upstream, proxy_region, resolve_dns, region)
        ret.append(lb_endpoints(**kw_args))
    return ret


def lb_endpoints(addrs: list, port: int, region: str, zone: str = None) -> dict:
    """
    Creates an envoy endpoint.LbEndpoints proto

    :param addrs:  The IP addresses or hostname(s) of the upstream.
    :param port:   The port that the upstream should be accessed on.
    :param region: The region of the upstream.
    :param zone:   The region of the proxy asking for the endpoint configuration.
    """
    node_priorities = priority_mapping.get(zone, {})
    priority = node_priorities.get(region, 10)
    return {
        'priority': priority,
        'locality': {'zone': region},
        'lb_endpoints': [{
            'endpoint': {
                'address': {
                    'socket_address': {
                        'address': addr,
                        'port_value': port
                    }
                }
            }
        } for addr in addrs]
    }



from fastapi.exceptions import HTTPException
from sovereign import config
from sovereign.statistics import stats
from sovereign.schemas import DiscoveryRequest
from sovereign.utils.crypto import decrypt, KEY_AVAILABLE, InvalidToken


def validate_authentication_string(s: str):
    try:
        password = decrypt(s)
    except Exception:
        stats.increment('discovery.auth.failed')
        raise

    if password in config.passwords:
        stats.increment('discovery.auth.success')
        return True
    stats.increment('discovery.auth.failed')
    return False


def authenticate(request: DiscoveryRequest):
    if not config.auth_enabled:
        return
    if not KEY_AVAILABLE:
        raise RuntimeError(
            'No Fernet key loaded, and auth is enabled. '
            'A fernet key must be provided via SOVEREIGN_ENCRYPTION_KEY. '
            'See https://vsyrakis.bitbucket.io/sovereign/docs/html/guides/encryption.html '
            'for more details'
        )
    try:
        encrypted_auth = request.node.metadata['auth']
        with stats.timed('discovery.auth.ms'):
            assert validate_authentication_string(encrypted_auth)
    except KeyError:
        raise HTTPException(status_code=401, detail=f'Discovery request from {request.node.id} is missing auth field')
    except (InvalidToken, AssertionError):
        raise HTTPException(status_code=401, detail='The authentication provided was invalid')
    except Exception as e:
        description = getattr(e, 'detail', 'Unknown')
        raise HTTPException(status_code=400, detail=f'The authentication provided was malformed [Reason: {description}]')



import zlib


def compute_hash(*args) -> str:
    """
    Creates a 'version hash' to be used in envoy Discovery Responses.
    """
    data: bytes = repr(args).encode()
    version_info = zlib.adler32(data) & 0xffffffff  # same numeric value across all py versions & platforms
    return str(version_info)



import random
from fastapi.routing import APIRouter
from fastapi.responses import PlainTextResponse
from sovereign import XDS_TEMPLATES, __versionstr__
from sovereign import discovery
from sovereign.sources import get_instances_for_node, extract_node_key
from sovereign.utils.mock import mock_discovery_request


router = APIRouter()


@router.get('/healthcheck', summary='Healthcheck (Does the server respond to HTTP?)')
async def health_check():
    return PlainTextResponse('OK')


@router.get('/deepcheck', summary='Deepcheck (Can the server render a random template?)')
async def deep_check():
    template = random.choice(
        list(XDS_TEMPLATES['default'].keys())
    )
    await discovery.response(
        mock_discovery_request(),
        xds_type=template
    )
    node = mock_discovery_request().node
    get_instances_for_node(node_value=extract_node_key(node))
    return PlainTextResponse(f'Rendered {template} OK')


@router.get('/version', summary='Display the current version of Sovereign')
async def version_check():
    return PlainTextResponse(f'Sovereign {__versionstr__}')



import sys
import schedule
from sovereign import config, XdsTemplate
from sovereign.config_loader import load
from sovereign.schemas import DiscoveryRequest
from sovereign.sources import extract_node_key, get_instances_for_node
from sovereign.statistics import stats
from sovereign.utils import crypto
from sovereign.utils.crypto import disabled_suite

template_context = {
    'crypto': crypto
}


def safe_context(request: DiscoveryRequest, template: XdsTemplate) -> dict:
    ret = build_template_context(
        node_value=extract_node_key(request.node),
        template=template,
    )
    # If the discovery request came from a mock, it will
    # typically contain this metadata key.
    # This means we should prevent any decryptable data
    # from ending up in the response.
    if request.hide_private_keys:
        ret['crypto'] = disabled_suite
    return ret


def build_template_context(node_value: str, template: XdsTemplate):
    """
    Creates context variables to be passed into either a jinja template,
    or as kwargs to a python template.
    """
    matches = get_instances_for_node(node_value=node_value)
    context = {**template_context}

    for scope, instances in matches.scopes.items():
        if scope == 'default':
            context['instances'] = instances
        else:
            context[scope] = instances

    if not template.is_python_source:
        for variable in list(context):
            if variable in template.jinja_variables():
                continue
            context.pop(variable, None)

    stats.set('discovery.context.bytes', sys.getsizeof(context))
    return context


def template_context_refresh():
    """ Modifies template_context in-place with new values """
    for k, v in config.template_context.items():
        template_context[k] = load(v)


# Initial setup
template_context_refresh()

if __name__ != '__main__' and config.refresh_context:  # pragma: no cover
    # This runs if the code was imported, as opposed to run directly
    schedule.every(config.context_refresh_rate).seconds.do(template_context_refresh)



"""
Inline Source
-------------

Example configuration (YAML):

.. code-block:: yaml

   sources:
     - type: inline
       config:
         instances:
           - instance_id: my_service
             service_clusters:
               - P2
             parameters:
               upstream_address:
                 - address: service.domain.com
                   region: us-east-1
             plan_id: 7d57270a-0348-58d3-829d-447a98fe98d5
"""
from sovereign.sources.lib import Source


class Inline(Source):
    def __init__(self, config, scope='default'):
        super(Inline, self).__init__(config, scope)
        try:
            self.instances = config['instances']
        except KeyError:
            raise KeyError('Inline source config must contain "instances"')

    def get(self):
        """ Returns the data passed via configuration """
        return self.instances



"""
Sources are used to retrieve arbitrary data that can then be represented
in templates as envoy configuration.

A new custom source can be added by subclassing the Source class and
adding an entry point using setuptools.

`todo add guide for adding entry points`

When the control plane receives a request for configuration from an envoy
proxy, it first runs through all the configured sources and combines the
data received from them into a single list, to be used within templates.

The results are cached for a configurable number of seconds to allow several proxies to poll
at the same time, receiving data that is consistent with each other.
"""
import schedule
import traceback
from functools import partial
from glom import glom, PathAccessError
from copy import deepcopy
from typing import Iterable, Any
from pkg_resources import iter_entry_points
from sovereign import config
from sovereign.modifiers import modify_sources_in_place
from sovereign.middlewares import get_request_id
from sovereign.statistics import stats
from sovereign.schemas import ConfiguredSource, SourceMetadata, SourceData, MemoizedTemplates
from sovereign.logs import LOG
from sovereign.sources.lib import Source
from sovereign.decorators import memoize

source_entry_points = iter_entry_points('sovereign.sources')
mod_entry_points = iter_entry_points('sovereign.modifiers')
gmod_entry_points = iter_entry_points('sovereign.global_modifiers')

memoized_templates = MemoizedTemplates()
source_metadata = SourceMetadata()
_source_data = SourceData()
_source_reference = {e.name: e.load() for e in source_entry_points}

_modifiers = dict()
_global_modifiers = dict()


if not _source_reference:
    raise RuntimeError('No sources available.')


def is_debug_request(v):
    return v == '' and config.debug_enabled


def is_wildcard(v):
    return v in [['*'], '*', ('*',)]


def contains(container, item):
    if isinstance(container, Iterable):
        return item in container


def load_modifiers(entry_points, configured_modifiers) -> dict:
    ret = dict()
    for entry_point in entry_points:
        if entry_point.name in configured_modifiers:
            ret[entry_point.name] = entry_point.load()
    return ret


def apply_modifications(source_data):
    """ 
    Wraps modify_sources_in_place so that modifier entry points 
    can be lazily loaded at runtime, only when modifications are
    required/need to be executed.
    """
    if not len(_modifiers):
        mods = load_modifiers(mod_entry_points, list(config.modifiers))
        _modifiers.update(mods)
    if not len(_global_modifiers):
        gmods = load_modifiers(gmod_entry_points, list(config.global_modifiers))
        _global_modifiers.update(gmods)
    return modify_sources_in_place(
        source_data, 
        _global_modifiers.values(), 
        _modifiers.values()
    )


def setup_sources(configured_source: ConfiguredSource) -> Source:
    """
    Takes Sources from config and turns them into Python objects, ready to use.
    """
    source_class = _source_reference[configured_source.type]
    source = source_class(
        config=configured_source.config,
        scope=configured_source.scope,
    )
    source.setup()
    return source


def sources_refresh():
    """
    All source data is stored in ``sovereign.sources._source_data``.
    Since the variable is outside this functions scope, we can only make
    in-place modifications to it via its methods.

    This function retrieves all sources, puts them into a temporary list,
    and then clears and re-fills ``_source_data`` with the new data.

    The process is done in two steps to avoid ``_source_data`` being empty
    for any significant amount of time.
    """
    stats.increment('sources.attempt')
    try:
        new_source_data = SourceData()
        for configured_source in config.sources:
            source = setup_sources(configured_source)
            new_source_data.scopes[source.scope].extend(source.get())
    except Exception as e:
        LOG.error(
            'Error while refreshing sources',
            traceback=[line for line in traceback.format_exc().split('\n')],
            error=e.__class__.__name__,
            detail=getattr(e, 'detail', '-'),
            request_id=get_request_id()
        )
        stats.increment('sources.error')
        raise

    if new_source_data == _source_data:
        stats.increment('sources.unchanged')
        source_metadata.update_date()
        return
    else:
        stats.increment('sources.refreshed')
        memoized_templates.purge()

    _source_data.scopes.clear()
    _source_data.scopes.update(new_source_data.scopes)
    source_metadata.update_date()
    source_metadata.update_count([
        instance
        for scope in _source_data.scopes.values()
        for instance in scope
    ])


@memoize(config.sources_refresh_rate * 0.8)
def read_sources() -> SourceData:
    """
    Returns a copy of source data in order to ensure it is not
    modified outside of the refresh function
    """
    return deepcopy(_source_data)


def get_instances_for_node(node_value: Any, modify=True, sources: SourceData = None, discovery_type: str = 'default') -> SourceData:
    """
    Checks a node against all sources, using the node_match_key and source_match_key
    to determine if the node should receive the source in its configuration.

    :param discovery_type: type of XDS request, used to determine if a scoped source should be evaluated
    :param node_value: value from the node portion of the envoy discovery request
    :param modify: switch to enable or disable modifications via Modifiers
    :param sources: the data sources to match the node against
    """
    if source_metadata.is_stale:
        # Log/emit metric and manually refresh sources.
        stats.increment('sources.stale')
        LOG.warn(
            'Sources have not been refreshed in 2 minutes',
            last_update=source_metadata.updated.isoformat(),
            instance_count=source_metadata.count
        )
        sources_refresh()

    if sources is None:
        sources: SourceData = read_sources()

    ret = SourceData()
    for scope, instances in sources.scopes.items():
        if config.node_matching is False:
            ret.scopes[scope] = instances
            continue

        for instance in instances:
            source_value = extract_source_key(instance)

            # If a single expression evaluates true, the remaining are not evaluated/executed.
            # This saves (a small amount of) computation, which helps when the server starts
            # to receive thousands of requests. The list has been ordered descending by what
            # we think will more commonly be true.
            match = (
                    contains(source_value, node_value)
                    or node_value == source_value
                    or is_wildcard(node_value)
                    or is_wildcard(source_value)
                    or is_debug_request(node_value)
            )
            if match:
                ret.scopes[scope].append(instance)
    if modify:
        return apply_modifications(ret)
    return ret


def extract_node_key(node):
    if '.' not in config.node_match_key:
        # key is not nested, don't need glom
        node_value = getattr(node, config.node_match_key)
    else:
        try:
            node_value = glom(node, config.node_match_key)
        except PathAccessError:
            raise RuntimeError(
                f'Failed to find key "{config.node_match_key}" in discoveryRequest({node}).\n'
                f'See the docs for more info: '
                f'https://vsyrakis.bitbucket.io/sovereign/docs/html/guides/node_matching.html'
            )
    return node_value


def extract_source_key(source):
    if '.' not in config.source_match_key:
        # key is not nested, don't need glom
        source_value = source[config.source_match_key]
    else:
        try:
            source_value = glom(source, config.source_match_key)
        except PathAccessError:
            raise RuntimeError(
                f'Failed to find key "{config.source_match_key}" in instance({source}).\n'
                f'See the docs for more info: '
                f'https://vsyrakis.bitbucket.io/sovereign/docs/html/guides/node_matching.html'
            )
    return source_value


def available_service_clusters():  # TODO: should be renamed since source value can be anything, not just service cluster
    """
    Checks for all match keys present in existing sources and adds them to a list

    A dict is used instead of a set because dicts cannot have duplicate keys, and
    have ordering since python 3.6
    """
    ret = dict()
    ret['*'] = None
    for scope, instances in read_sources().scopes.items():
        if config.node_matching is False:
            break

        for instance in instances:
            source_value = glom(instance, config.source_match_key)
            if isinstance(source_value, Iterable):
                for item in source_value:
                    ret[item] = None
                continue
            ret[source_value] = None
    return list(ret.keys())


if __name__ != '__main__':
    schedule.every(config.sources_refresh_rate).seconds.do(sources_refresh)



from random import randint
from sovereign.schemas import DiscoveryRequest, Node, Locality


def mock_discovery_request(service_cluster=None, resource_names=None, region='none', version='1.11.1', metadata=None) -> DiscoveryRequest:
    if not isinstance(metadata, dict):
        metadata = dict()
    return DiscoveryRequest(
        node=Node(
            id='mock',
            cluster=service_cluster or '',
            build_version=f'e5f864a82d4f27110359daa2fbdcb12d99e415b9/{version}/Clean/RELEASE',
            locality=Locality(
                zone=region
            ),
            metadata={
                **metadata,
            }
        ),
        version_info=str(randint(100000, 1000000000)),
        resource_names=resource_names or [],
        hide_private_keys=True,
    )



from fastapi import Body, Header
from fastapi.routing import APIRouter
from fastapi.responses import Response
from sovereign.logs import add_log_context
from sovereign import discovery
from sovereign.sources import memoized_templates as nodes
from sovereign.schemas import DiscoveryRequest, DiscoveryResponse, ProcessedTemplate
from sovereign.statistics import stats
from sovereign.utils.auth import authenticate

router = APIRouter()

type_urls = {
    'v2': {
        'listeners': 'type.googleapis.com/envoy.api.v2.Listener',
        'clusters': 'type.googleapis.com/envoy.api.v2.Cluster',
        'endpoints': 'type.googleapis.com/envoy.api.v2.ClusterLoadAssignment',
        'secrets': 'type.googleapis.com/envoy.api.v2.auth.Secret',
        'routes': 'type.googleapis.com/envoy.api.v2.RouteConfiguration',
        'scoped-routes': 'type.googleapis.com/envoy.api.v2.ScopedRouteConfiguration',
    },
    'v3': {
        'listeners': 'type.googleapis.com/envoy.config.listener.v3.Listener',
        'clusters': 'type.googleapis.com/envoy.config.cluster.v3.Cluster',
        'routes': 'type.googleapis.com/envoy.config.route.v3.RouteConfiguration',
        'scoped-routes': 'type.googleapis.com/envoy.config.route.v3.ScopedRouteConfiguration',
    },
}


def response_headers(discovery_request, response, xds):
    return {
        'X-Sovereign-Client-Build': discovery_request.envoy_version,
        'X-Sovereign-Client-Version': discovery_request.version_info,
        'X-Sovereign-Requested-Resources': ','.join(discovery_request.resource_names) or 'all',
        'X-Sovereign-Requested-Type': xds,
        'X-Sovereign-Response-Version': response.version_info
    }


@router.post(
    '/{version}/discovery:{xds_type}',
    summary='Envoy Discovery Service Endpoint',
    response_model=DiscoveryResponse,
    responses={
        200: {'description': 'New resources provided'},
        304: {'description': 'Resources are up-to-date'},
        404: {'description': 'No resources found'},
    }
)
async def discovery_response(
        version: str,
        xds_type: discovery.DiscoveryTypes,
        discovery_request: DiscoveryRequest = Body(None),
        host: str = Header('no_host_provided'),
):
    xds = xds_type.value
    discovery_request.desired_controlplane = host
    add_log_context(
        resource_names=discovery_request.resource_names,
        envoy_ver=discovery_request.envoy_version
    )
    response = await perform_discovery(discovery_request, version, xds, skip_auth=False)
    headers = response_headers(discovery_request, response, xds)
    if response.version_info == discovery_request.version_info:
        return not_modified(headers)
    elif len(response.resources) == 0:
        return Response(status_code=404, headers=headers)
    elif response.version_info != discovery_request.version_info:
        return Response(response.rendered, headers=headers, media_type='application/json')
    return Response(content='Resources could not be determined', status_code=500)


async def perform_discovery(req, api_version, xds, skip_auth=False) -> ProcessedTemplate:
    if not skip_auth:
        authenticate(req)
    try:
        type_url = type_urls[api_version][xds]
        req.type_url = type_url
    except TypeError:
        pass
    # Only run this block if the envoy proxy flags that it wants this behavior
    if req.node.metadata.get('enable_beta_caching'):
        # Attempt to retrieve cached data
        cached_data = nodes.get_node(req.uid, xds)
        if cached_data:
            stats.increment(f'discovery.{xds}.cache_hit')
            return cached_data
        else:
            # Perform normal discovery and then add it to the cache
            response = await discovery.response(req, xds)
            stats.increment(f'discovery.{xds}.cache_miss')
            nodes.add_node(uid=req.uid, xds_type=xds, template=response)
            return response
    return await discovery.response(req, xds)


def not_modified(headers):
    return Response(status_code=304, headers=headers)



"""
Discovery
---------

Functions used to render and return discovery responses to Envoy proxies.

The templates are configurable. `todo See ref:Configuration#Templates`
"""
import yaml
from yaml.parser import ParserError, ScannerError
from enum import Enum
from typing import Union
from collections import namedtuple
from starlette.exceptions import HTTPException
from sovereign import XDS_TEMPLATES, config
from sovereign.utils.version_info import compute_hash
from sovereign.logs import LOG
from sovereign.context import safe_context
from sovereign.schemas import XdsTemplate, DiscoveryRequest, ProcessedTemplate

try:
    import sentry_sdk
except ImportError:
    sentry_sdk = None

try:
    default_templates = XDS_TEMPLATES['default']
except KeyError:
    raise KeyError(
        'Your configuration should contain default templates. For more details, see '
        'https://vsyrakis.bitbucket.io/sovereign/docs/html/guides/tutorial.html#create-templates '
    )

# Create an enum that bases all the available discovery types off what has been configured
discovery_types = (_type for _type in sorted(XDS_TEMPLATES['__any__'].keys()))
# noinspection PyArgumentList
all_types = {t: t for t in discovery_types}
DiscoveryTypes = Enum('DiscoveryTypes', all_types)

NotModified = namedtuple('NotModified', ['version_info', 'resources'])


def select_template(request: DiscoveryRequest, discovery_type: DiscoveryTypes, templates=None) -> XdsTemplate:
    if templates is None:
        templates = XDS_TEMPLATES
    version = request.envoy_version
    selection = 'default'
    for v in templates.keys():
        if version.startswith(v):
            selection = v
    selected_version = templates[selection]
    try:
        return selected_version[discovery_type]
    except KeyError:
        raise KeyError(f'Unable to get {discovery_type} for template version "{selection}". Envoy client version: {version}')


async def response(request: DiscoveryRequest, xds_type: DiscoveryTypes) -> Union[ProcessedTemplate, NotModified]:
    """
    A Discovery **Request** typically looks something like:

    .. code-block:: json

        {
            "version_info": "0",
            "node": {
                "cluster": "T1",
                "build_version": "<revision hash>/<version>/Clean/RELEASE",
                "metadata": {
                    "auth": "..."
                }
            }
        }

    When we receive this, we give the client the latest configuration via a
    Discovery **Response** that looks something like this:

    .. code-block:: json

        {
            "version_info": "abcdef1234567890",
            "resources": []
        }

    The version_info is derived from :func:`sovereign.discovery.version_hash`

    :param request: An envoy Discovery Request
    :param xds_type: what type of XDS template to use when rendering
    :return: An envoy Discovery Response
    """
    template: XdsTemplate = select_template(request, xds_type)
    context: dict = safe_context(request, template)

    config_version = None
    if config.cache_strategy == 'context':
        config_version = compute_hash(
            context,
            template.checksum,
            request.node.common,
            request.resource_names,
            request.desired_controlplane
        )
        if config_version == request.version_info:
            return NotModified(version_info=config_version, resources=[])

    context = dict(
        discovery_request=request,
        host_header=request.desired_controlplane,
        resource_names=request.resources,
        **context
    )
    content = await template(**context)
    if not template.is_python_source:
        content = deserialize_config(content)

    if config.cache_strategy == 'content':
        config_version = compute_hash(content)
        if config_version == request.version_info:
            return NotModified(version_info=config_version, resources=[])

    resources = filter_resources(content['resources'], request.resources)
    return ProcessedTemplate(resources=resources, type_url=request.type_url, version_info=config_version)


def deserialize_config(content):
    try:
        envoy_configuration = yaml.safe_load(content)
    except (ParserError, ScannerError) as e:
        LOG.msg(
            error=repr(e),
            context=e.context,
            context_mark=e.context_mark,
            note=e.note,
            problem=e.problem,
            problem_mark=e.problem_mark,
        )

        if config.sentry_dsn and sentry_sdk:
            sentry_sdk.capture_exception(e)

        raise HTTPException(
            status_code=500,
            detail='Failed to load configuration, there may be '
                   'a syntax error in the configured templates. '
                   'Please check Sentry if you have configured Sentry DSN'

        )
    return envoy_configuration


def filter_resources(generated, requested):
    """
    If Envoy specifically requested a resource, this removes everything
    that does not match the name of the resource.
    If Envoy did not specifically request anything, every resource is retained.
    """
    return [
        resource
        for resource in generated
        if resource_name(resource) in requested
    ]


def resource_name(resource):
    return resource.get('name') or resource['cluster_name']



from collections import defaultdict
from fastapi import APIRouter, Query, Path, Cookie
from fastapi.encoders import jsonable_encoder
from fastapi.requests import Request
from fastapi.responses import RedirectResponse, JSONResponse, Response
from sovereign import html_templates, XDS_TEMPLATES
from sovereign.discovery import DiscoveryTypes
from sovereign import json_response_class
from sovereign.sources import available_service_clusters, source_metadata
from sovereign.utils.mock import mock_discovery_request
from sovereign.views.discovery import perform_discovery

router = APIRouter()

all_types = [t.value for t in DiscoveryTypes]


@router.get('/')
async def ui_main(request: Request):
    try:
        return html_templates.TemplateResponse(
            name='base.html',
            media_type='text/html',
            context={
                'request': request,
                'all_types': all_types,
                'last_update': str(source_metadata),
            })
    except IndexError:
        return html_templates.TemplateResponse(
            name='err.html',
            media_type='text/html',
            context={
                'request': request,
                'title': 'No resource types configured',
                'message': 'A template should be defined for every resource '
                           'type that you want your envoy proxies to discover.',
                'doc_link': 'https://vsyrakis.bitbucket.io/sovereign/docs/tutorial/templates/',
            })


@router.get(
    '/set-version',
    summary='Filter the UI by a certain Envoy Version (stores a Cookie)'
)
async def set_envoy_version(
        request: Request,
        version: str = Query('__any__', title='The clients envoy version to emulate in this XDS request'),
):
    url = request.headers.get('Referer', '/ui')
    response = RedirectResponse(url=url)
    response.set_cookie(key='envoy_version', value=version)
    return response


@router.get(
    '/set-service-cluster',
    summary='Filter the UI by a certain service cluster (stores a Cookie)'
)
async def set_service_cluster(
        request: Request,
        service_cluster: str = Query('__any__', title='The clients envoy version to emulate in this XDS request'),
):
    url = request.headers.get('Referer', '/ui')
    response = RedirectResponse(url=url)
    response.set_cookie(key='service_cluster', value=service_cluster)
    return response


@router.get(
    '/resources/{xds_type}',
    summary='List available resources for a given xDS type'
)
async def resources(
        request: Request,
        xds_type: DiscoveryTypes = Path('clusters', title='xDS type', description='The type of request'),
        region: str = Query(None, title='The clients region to emulate in this XDS request'),
        api_version: str = Query('v2', title='The desired Envoy API version'),
        service_cluster: str = Cookie('*', title='The clients service cluster to emulate in this XDS request'),
        envoy_version: str = Cookie('__any__', title='The clients envoy version to emulate in this XDS request')
):
    ret = defaultdict(list)
    try:
        response = await perform_discovery(
            req=mock_discovery_request(
                service_cluster=service_cluster,
                resource_names=[],
                version=envoy_version,
                region=region
            ),
            api_version=api_version,
            xds=xds_type.value,
            skip_auth=True,
        )
    except KeyError:
        ret['resources'] = []
    else:
        ret['resources'] += response.resources
    return html_templates.TemplateResponse(
        name='resources.html',
        media_type='text/html',
        context={
            'resources': ret['resources'],
            'request': request,
            'resource_type': xds_type.value,
            'all_types': all_types,
            'version': envoy_version,
            'available_versions': list(XDS_TEMPLATES.keys()),
            'service_cluster': service_cluster,
            'available_service_clusters': available_service_clusters(),
            'last_update': str(source_metadata),
        })


@router.get(
    '/resources/{xds_type}/{resource_name}',
    summary='Return JSON representation of a resource'
)
async def resource(
        xds_type: DiscoveryTypes = Path('clusters', title='xDS type', description='The type of request'),
        resource_name: str = Path(..., title='Name of the resource to view'),
        region: str = Query(None, title='The clients region to emulate in this XDS request'),
        api_version: str = Query('v2', title='The desired Envoy API version'),
        service_cluster: str = Cookie('*', title='The clients service cluster to emulate in this XDS request'),
        envoy_version: str = Cookie('__any__', title='The clients envoy version to emulate in this XDS request')
):
    response = await perform_discovery(
        req=mock_discovery_request(
            service_cluster=service_cluster,
            resource_names=[resource_name],
            version=envoy_version,
            region=region
        ),
        api_version=api_version,
        xds=xds_type.value,
        skip_auth=True,
    )
    return Response(response.rendered, media_type='application/json')


@router.get(
    '/resources/routes/{route_configuration}/{virtual_host}',
    summary='Return JSON representation of Virtual Hosts'
)
async def virtual_hosts(
        route_configuration: str = Path(..., title='Name of the route configuration'),
        virtual_host: str = Path(..., title='Name of the resource to view'),
        region: str = Query(None, title='The clients region to emulate in this XDS request'),
        api_version: str = Query('v2', title='The desired Envoy API version'),
        service_cluster: str = Cookie('*', title='The clients service cluster to emulate in this XDS request'),
        envoy_version: str = Cookie('__any__', title='The clients envoy version to emulate in this XDS request')
):
    response = await perform_discovery(
        req=mock_discovery_request(
            service_cluster=service_cluster,
            resource_names=[route_configuration],
            version=envoy_version,
            region=region
        ),
        api_version=api_version,
        xds='routes',
        skip_auth=True,
    )
    route_configs = [
        resource_
        for resource_ in response.resources
        if resource_['name'] == route_configuration
    ]
    for route_config in route_configs:
        for vhost in route_config['virtual_hosts']:
            if vhost['name'] == virtual_host:
                safe_response = jsonable_encoder(vhost)
                try:
                    return json_response_class(content=safe_response)
                except TypeError:
                    return JSONResponse(content=safe_response)
        break



import gunicorn.app.base
from sovereign import asgi_config
from sovereign.app import app


class StandaloneApplication(gunicorn.app.base.BaseApplication):
    def __init__(self, application, options=None):
        self.options = options or {}
        self.application = application
        super().__init__()

    def load_config(self):
        for key, value in self.options.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        return self.application


def main():
    asgi = StandaloneApplication(
        application=app,
        options=asgi_config.as_gunicorn_conf()
    )
    asgi.run()


if __name__ == '__main__':
    main()
